package org.pHash;

import java.io.*;

/**
 * Service provider for pHash.
 *
 * TODO
 * - getters for default options
 *
 * @author Christoph Zauner
 */
public class pHash
{

	/*** begin JNI bindings ***/

	/*
	 * Ideally the API should provide the following functions for
	 * each perceptual hash function:
	 *
	 * 1. function to get hash of a file (e.g. getBMBImageHash)
	 * 2. function to compare two hashes (e.g. getBMBImageHashDistance, method overloading)
	 * 3. function to compare two files  (e.g. getBMBImageHashDistance, method overloading)
	 */

	// library management
	private native static void pHashInit();
	private native static void cleanup();

	// audio
	private native static AudioHash audioHash(String file);
	// image
	private native static BMBImageHash    bmbImageHash(String file, int method, int presetSizeX, int blockSizeX);
	private native static double          bmbImagesDistance(String file1, String file2, int presetSizeX, int blockSizeX); // TODO remove ?
	private native static DCTImageHash    dctImageHash(String file);
	private native static MHImageHash     mhImageHash(String file, float alpha, float level);
	private native static RadialImageHash radialImageHash(String file, double sigma, double gamma, int angles); // TODO implement
	private native static double          radialImagesDistance(String file1, String file2, double sigma, double gamma, int angles); // TODO remove ?
	// text
	private native static TextHash textHash(String file);
	// video
	private native static VideoHash videoHash(String file);

	// audio
	private native static double audioDistance(AudioHash hash1, AudioHash hash2);
	// image
	private native static double imageDistance(ImageHash hash1, ImageHash hash2); // used for all the image hashes
	// text
	private native static int textDistance(TextHash txtHash1, TextHash txtHash2);
	// video
	private native static double videoDistance(VideoHash hash1, VideoHash hash2, int threshold);

	/*** end JNI bindings ***/

	/*** begin hash function arguments ***/
	private static int bmb_method = 1;
	private static int bmb_presetSizeX = 256;
	private static int bmb_blockSizeX = 16;

	private static float mh_alpha = 2;
	private static float mh_level = 1;

	private static double radial_sigma = 1;
	private static double radial_gamma = 1;
	private static int radial_angles = 180;
	/*** end hash function arguments ***/

	static {
		System.loadLibrary("pHash-jni");
		pHashInit();
	}

	// BMB image hash
	
	public static BMBImageHash getBMBImageHash(File file) {
		BMBImageHash imHash = bmbImageHash(file.toString(), bmb_method, bmb_presetSizeX, bmb_blockSizeX);
		return imHash;
	}

	public static double getBMBImageHashDistance(BMBImageHash hashA, BMBImageHash hashB) {
		return(imageDistance(hashA, hashB));
	}

	public static double getBMBImagesDistance(String file1, String file2) {
		BMBImageHash imHash1 = bmbImageHash(file1, bmb_method, bmb_presetSizeX, bmb_blockSizeX);
		BMBImageHash imHash2 = bmbImageHash(file2, bmb_method, bmb_presetSizeX, bmb_blockSizeX);
		return(imageDistance(imHash1, imHash2));
	}

	public static void setBMBImageHashArgs(int method, int presetSizeX, int blockSizeX) {
		bmb_method = method;
		bmb_presetSizeX = presetSizeX;
		bmb_blockSizeX = blockSizeX;
	}

	// Radial image hash

	public static RadialImageHash getRadialImageHash(File file) {
		RadialImageHash imHash = radialImageHash(file.toString(), radial_sigma, radial_gamma, radial_angles);
		return imHash;
	}

	public static double getRadialImageHashDistance(RadialImageHash hashA, RadialImageHash hashB) {
		return(imageDistance(hashA, hashB));
	}

	public static double getRadialImagesDistance(String file1, String file2) {
		return(radialImagesDistance(file1, file2, radial_sigma, radial_gamma, radial_angles));
//		RadialImageHash imHash1 = radialImageHash(file1);
//		RadialImageHash imHash2 = radialImageHash(file2);
//		return(imageDistance(imHash1, imHash2));
	}

	public static void setRadialImageHashArgs(double sigma, double gamma, int angles) {
		radial_sigma = sigma;
		radial_gamma = gamma;
		radial_angles = angles;
	}

	// DCT image hash

	public static DCTImageHash getDCTImageHash(File file) {
		DCTImageHash imHash = dctImageHash(file.toString());
		return imHash;
	}

	public static double getDCTImageHashDistance(DCTImageHash hashA, DCTImageHash hashB) {
		return(imageDistance(hashA, hashB));
	}

	public static double getDCTImageHashDistance(String file1, String file2) {
		DCTImageHash imHash1 = dctImageHash(file1);
		DCTImageHash imHash2 = dctImageHash(file2);
		return(imageDistance(imHash1,imHash2));
	}

	// MH image hash

	public static MHImageHash getMHImageHash(File file) {
		MHImageHash imHash = mhImageHash(file.toString(), mh_alpha, mh_level);
		return imHash;
	}

	public static double getMHImageHashDistance(MHImageHash hashA, MHImageHash hashB) {
		return(imageDistance(hashA, hashB));
	}
	
	public static double getMHImageHashDistance(String file1, String file2) {
		MHImageHash imHash1 = mhImageHash(file1, mh_alpha, mh_level);
		MHImageHash imHash2 = mhImageHash(file2, mh_alpha, mh_level);
		return(imageDistance(imHash1,imHash2));
	}

	public static void setMHImageHashArgs(float alpha, float level) {
		mh_alpha = alpha;
		mh_level = level;
	}

	/**
	 * Just a test function to test the JNI calls.
	 *
	 * @param args
	 */
	public static void main(String args[]) {
			int i = 0;

			/* MVP Hash storage */

			if(args[i].equals("-mvp"))
			{
					MVPTree mvp = new MVPTree("mvp");
					MHImageHash[] hashes = getMHImageHashes(args[1]);
					boolean result = mvp.create(hashes);
					if(result)
					{
						System.out.println("Successfully created MVP tree");
						Hash[] results = mvp.query(hashes[0], 100, 20, 30);
						if(results != null && results.length > 0)
						{
						System.out.println("Query found " + results.length + " results");
						for(int j = 0; j < results.length; ++j)
							System.out.println("File: " + results[j].filename);
						}

						MHImageHash[] newHashes = getMHImageHashes(args[2]);

						boolean added = mvp.add(newHashes);
						if(added)
						{
							System.out.println("Hashes added successfully.");
							Hash[] foundHashes = mvp.query(newHashes[0], 100, 20, 30);
							if(foundHashes != null && foundHashes.length > 0)
							{
								System.out.println("Found newly added hash.");
							}
						}
					} else
						System.out.println("Creating tree failed");

			}

			/* Misc. hash function */

			else if(args[i].equals("-a"))
			{
				AudioHash audioHash1 = audioHash(args[1]);
				AudioHash audioHash2 = audioHash(args[2]);
				System.out.println("cs = " + audioDistance(audioHash1,audioHash2));
			}
			else if(args[i].equals("-v"))
			{
				VideoHash vHash = videoHash(args[1]);
				VideoHash vHash2 = videoHash(args[2]);
				System.out.println(videoDistance(vHash,vHash2, 21));
			}
			else if(args[i].equals("-t"))
			{
				TextHash txtHash = textHash(args[1]);
				TextHash txtHash2 = textHash(args[2]);
				System.out.println(textDistance(txtHash,txtHash2));
			}

			/* DCT Image Hash */

			else if(args[i].equals("-dct"))
			{
				checkArgs(3, args.length);
				DCTImageHash imHash = dctImageHash(args[1]);
				DCTImageHash imHash2 = dctImageHash(args[2]);
				System.out.println("File 1: " + imHash.filename);
				System.out.println("File 2: " + imHash2.filename);

				System.out.println("Distance: " + imageDistance(imHash,imHash2));
			}
			else if(args[i].equals("-dctHash"))
			{
				checkArgs(2, args.length);
				DCTImageHash imHash = dctImageHash(args[1]);
				System.out.printf("DCT hash of image >%s< is >%s<\n", args[1], imHash);
			}

			/* MH Image Hash */

			else if(args[i].equals("-mh"))
			{
				checkArgs(3, args.length);
				MHImageHash imHash = mhImageHash(args[1], mh_alpha, mh_level);
				MHImageHash imHash2 = mhImageHash(args[2], mh_alpha, mh_level);
				System.out.println("File 1: " + imHash.filename);
				System.out.println("File 2: " + imHash2.filename);

				System.out.println("Distance: " + imageDistance(imHash,imHash2));
			} 
			else if(args[i].equals("-mhHash"))
			{
				checkArgs(2, args.length);
				MHImageHash imHash = mhImageHash(args[1], mh_alpha, mh_level);
				System.out.printf("MH hash of image >%s< is >%s<\n", args[1], imHash);
			}

			/* BMB Image Hash */

			else if(args[i].equals("-bmb"))
			{
				checkArgs(3, args.length);
				BMBImageHash imHash = bmbImageHash(args[1], bmb_method, bmb_presetSizeX, bmb_blockSizeX);
				BMBImageHash imHash2 = bmbImageHash(args[2], bmb_method, bmb_presetSizeX, bmb_blockSizeX);
				System.out.println("File 1: " + imHash.filename);
				System.out.println("File 2: " + imHash2.filename);

				System.out.println("Distance: " + imageDistance(imHash,imHash2));
			}
			else if(args[i].equals("-bmbHash"))
			{
				BMBImageHash imHash = bmbImageHash(args[1], bmb_method, bmb_presetSizeX, bmb_blockSizeX);
				System.out.printf("BMB hash of image >%s< is >%s<\n", args[1], imHash);
			}

			/* Radial Image Hash */

			else if(args[i].equals("-radial"))
			{
				checkArgs(3, args.length);
				RadialImageHash imHash = radialImageHash(args[1], radial_sigma, radial_gamma, radial_angles);
				RadialImageHash imHash2 = radialImageHash(args[2], radial_sigma, radial_gamma, radial_angles);
				System.out.println("File 1: " + imHash.filename);
				System.out.println("File 2: " + imHash2.filename);

				System.out.println("Distance: " + imageDistance(imHash,imHash2));
			}
			else if(args[i].equals("-radialHash"))
			{
				RadialImageHash imHash = radialImageHash(args[1], radial_sigma, radial_gamma, radial_angles);
				System.out.printf("Radial hash of image >%s< is >%s<\n", args[1], imHash);
			}

			pHash.cleanup();
	}

	/**
	 * Helper function for main().
	 * 
	 * @param d
	 * @return
	 */
	private static MHImageHash[] getMHImageHashes(String d) {
		File dir = new File(d);
		MHImageHash[] hashes = null;
		if(dir.isDirectory())
		{
			File[] files = dir.listFiles();
			hashes = new MHImageHash[files.length];
			for(int i = 0; i < files.length; ++i)
			{
				MHImageHash mh = mhImageHash(files[i].toString(), mh_alpha, mh_level);
				if(mh != null)
					hashes[i] = mh;
			}
		}
		return hashes;
	}

	private static void checkArgs(int actualArgs, int reqArgs) {
		if(actualArgs != reqArgs) {
			System.out.println("Error in pHash.java: invalid command line args");
			System.exit(1);
		}
	}

} // end of class
